"""Settings for the questions module."""


class Settings:
    """Settings for the questions module."""

    MAX_ANSWER_LENGTH = 2000
    MAX_EXPRESSION_CONSTRAINT_LENGTH = 1000
    MAX_NUM_OPTIONS = 200
    MIN_NUM_OPTIONS = 2
    MAX_OPTION_LENGTH = 1000
    MAX_QUESTION_LENGTH = 100000
